import logging
import logging.handlers
import time
from io import BytesIO
from queue import Queue

import requests

__all__ = ("StandaloneHandler", "ThreadedHandler")


class BaseHandler(logging.Handler):
    """⭐"""

    def __init__(self):
        """⭐"""
        super().__init__()

    def _request(self, method, **kwargs):
        """⭐"""
        try:
            response = self.session.post(self.url_template.format(method=method), **kwargs)
            response.raise_for_status()
        except Exception as exc:
            print("Can't send log telegram message: ", exc)

    def _send(self, text_content, **data):
        """⭐"""
        if len(text_content) < 4096:
            self._request("sendMessage", json={'text': text_content, **data})
        else:
            self._request(
                "sendDocument",
                data={"caption": text_content[:self.max_message_slice], **data},
                files={"document": (self.document_name, BytesIO(text_content.encode("UTF-8")), "text/plain")}
                )

    def _wait(self):
        """⭐"""

    def handle(self, record):
        """⭐"""
        record = self.format(record)
        self.emit(record)

    def emit(self, record):
        """⭐"""
        for chat_id in self.chat_ids:
            self._send(record, **{"chat_id": chat_id, **self.rq_data, **self.rq_params})
            self._wait()


class ThreadedSlaveHandler(BaseHandler):
    """⭐"""

    def __init__(
        self,
        token,
        chat_ids,
        disable_web_page_preview,
        disable_notification,
        parse_mode,
        max_message_slice,
        rps_limit,
        document_name,
        **rq_params
        ):
        """⭐"""
        self.chat_ids = chat_ids
        self.rq_params = rq_params
        self.max_message_slice = max_message_slice
        self.document_name = document_name
        self.rps_limit = rps_limit
        self.url_template = "https://api.telegram.org/bot{token}".format(token=token) + "/{method}"
        self.rq_data = {
            "disable_web_page_preview": disable_web_page_preview,
            "disable_notification": disable_notification,
            "parse_mode": parse_mode
            }
        self.session = requests.Session()
        super().__init__()

    def _wait(self):
        """⭐"""
        time.sleep(self.rps_limit)


class StandaloneHandler(BaseHandler):
    """⭐"""

    def __init__(
        self,
        token,
        chat_ids,
        disable_web_page_preview=False,
        disable_notification=False,
        parse_mode="html",
        max_message_slice=1024,
        rps_limit=0.03,
        document_name="log.txt",
        **rq_params
        ):
        """⭐"""
        self.chat_ids = chat_ids
        self.rq_params = rq_params
        self.max_message_slice = max_message_slice
        self.document_name = document_name
        self.rps_limit = rps_limit
        self.url_template = "https://api.telegram.org/bot{token}".format(token=token) + "/{method}"
        self.rq_data = {
            "disable_web_page_preview": disable_web_page_preview,
            "disable_notification": disable_notification,
            "parse_mode": parse_mode
            }
        self.session = requests.Session()
        super().__init__()


class ThreadedHandler(logging.handlers.QueueHandler):
    """⭐"""

    def __init__(
        self,
        token,
        chat_ids,
        disable_web_page_preview=False,
        disable_notification=False,
        parse_mode="html",
        max_message_slice=1024,
        rps_limit=0.03,
        document_name="log.txt",
        **rq_params
        ):
        """⭐"""
        queue: Queue = Queue()
        super().__init__(queue)
        handler = ThreadedSlaveHandler(
            token, chat_ids, disable_web_page_preview, disable_notification,
            parse_mode, max_message_slice, rps_limit, document_name, **rq_params
            )
        listener = logging.handlers.QueueListener(queue, handler)
        listener.start()



# -*- coding: utf-8 -*-
from asyncio import create_task, gather
from asyncio import sleep as asleep
from collections import Counter
from datetime import datetime, timedelta
from functools import reduce
from logging import getLogger
from uuid import UUID

from aiohttp import ClientSession
from sqlalchemy import text

from vpnflow.db import repositories as repo
from vpnflow.db.base import engine_mrzbn, engine_r, engine_w, run_session
from vpnflow.db.tools import (get_marzban_users, get_soon_expired_users,
                              get_telegram_expired_users, load_external_table)
from vpnflow.misc import slice_generator
from vpnflow.services._marzban import Marzban
from vpnflow.services._telegram import broadcast, parse_ref_param_start
from vpnflow.services.cache import DEFAULT_TTL, CacheRepository
from vpnflow.settings import settings
from vpnflow.web import schemas

logger = getLogger(__name__)


async def is_telegram_user_exists(telegram_id: int, engine=engine_r) -> bool:
    """⭐"""
    cache_result = await CacheRepository.get(f"tg_exists:{telegram_id}")
    if cache_result:
        return cache_result
    async with run_session(engine) as session:
        user = await repo.UsersTelegramRepository.find_one_or_none(session, id=telegram_id)
        if user:
            await CacheRepository.set(f"tg_exists:{telegram_id}", True, DEFAULT_TTL)
            return True
    return False


async def create_user_from_telegram(user_data, command_args, engine=engine_r):
    """⭐"""
    tg_user = schemas.UserTelegram(
        id=user_data.id, name=user_data.username, name_full=user_data.full_name,
        is_bot=user_data.is_bot, language_code=user_data.language_code,
        is_premium=user_data.is_premium
        )
    referral = parse_ref_param_start(command_args) if command_args else None
    if referral and referral.telegram_id == tg_user.id:
        referral = None
    async with run_session(engine, commit=True) as session:
        tg_user_created = await repo.UsersTelegramRepository.add(session, tg_user)
        if referral:
            referral_created = await repo.ReferralRepository.add(session, referral)
            user = repo.UsersRepository.schema(
                telegram_id=tg_user_created.id, referral_id=referral_created.id
                )
        else:
            user = repo.UsersRepository.schema(telegram_id=tg_user_created.id)
        user_created = await repo.UsersRepository.add(session, user)
        return user_created


async def create_user_marzban(days, traffic_limit, proxies, inbounds, retries=10):
    """⭐"""
    async with ClientSession() as session:
        headers = await Marzban.auth(session)
        for _ in range(retries):
            username = Marzban.usernames_generator()
            result = await Marzban.request(
                session, "GET", f"/users?limit=10&search={username}", **headers
                )
            # params: offset limit username search status sort
            if result.get("total", 0) == 0:
                result = await Marzban.request(
                    session, "POST", "/user",
                    json=schemas.MarzbanUserCreate(
                        username=username,
                        data_limit=2 ** 30 * traffic_limit,
                        proxies=proxies,
                        inbounds=inbounds,
                        expire=int((datetime.utcnow() + timedelta(days=days)).timestamp())
                        ).dict(), **headers
                    )
                return schemas.MarzbanResponse(**result)


async def update_user(user, engine=engine_w, **values):
    """⭐"""
    await clean_cache(f"user:{user.telegram_id}")
    async with run_session(engine) as session:
        user_updated = await repo.UsersRepository.update(
            session, {"id": user.id}, **values
            )
    return user_updated


async def get_user(telegram_id, engine=engine_r):
    """⭐"""
    cache_result = await CacheRepository.get(f"user:{telegram_id}")
    if cache_result:
        return cache_result
    async with run_session(engine) as session:
        user = await repo.UsersRepository.find_one_or_none(session, telegram_id=telegram_id)
        if user:
            await CacheRepository.set(f"user:{telegram_id}", user, DEFAULT_TTL)
        return user


async def get_user_balances(user, engine=engine_r):
    """⭐"""
    balances = []
    async with run_session(engine) as session:
        for balance in user.balances:
            user_balance = schemas.UserBalanceView.model_validate(balance)
            currency = await repo.PayPlanPriceCurrenciesRepository.find_one_or_none(session, code=balance.currency)
            if not currency:
                logger.warning(f"Can't find currency: {balance.currency}")
                continue
            user_balance.symbol = currency.symbol
            balances.append(user_balance.text)
    return " ".join(balances) if balances else "0"


async def update_user_marzban(marzban_username, days, **kw):
    """⭐"""
    await clean_cache(f"marzban:{marzban_username}")
    async with ClientSession() as session:
        headers = await Marzban.auth(session)

        result = await Marzban.request(session, "GET", f"/user/{marzban_username}", **headers)
        marzban_user = schemas.MarzbanResponse(**result)

        ts_expire = marzban_user.expire
        ts_current = datetime.utcnow().timestamp()
        ts = int(ts_current) if ts_expire < ts_current else ts_expire
        ts += days * 86400 # 24 * 60 * 60
        data = {"status": "active", "expire": ts, **kw}

        result = await Marzban.request(session, "PUT", f"/user/{marzban_username}", json=data, **headers)
        return schemas.MarzbanResponse(**result)


async def get_user_marzban(username):
    """⭐"""
    cache_result = await CacheRepository.get(f"marzban:{username}")
    if cache_result:
        return cache_result
    async with ClientSession() as session:
        headers = await Marzban.auth(session)
        result = await Marzban.request(session, "GET", f"/user/{username}", **headers)
        result = schemas.MarzbanResponse(**result)
        await CacheRepository.set(f"marzban:{username}", result, DEFAULT_TTL)
        return result


async def get_pay_plans(engine=engine_r):
    """⭐"""
    async with run_session(engine) as session:
        for pay_plan in await repo.PayPlansRepository.find_all(session, is_active=True):
            # for pay_plan_price in await pay_plan.awaitable_attrs.prices:
            yield repo.PayPlansRepository.view.model_validate(pay_plan)


async def get_pay_plan(_id: int, engine=engine_r):
    """⭐"""
    _id = int(_id) # callback data str type
    cache_result = await CacheRepository.get(f"payplan:{_id}")
    if cache_result:
        return cache_result
    async with run_session(engine) as session:
        pay_plan = await repo.PayPlansRepository.find_one_or_none(session, **{"id": _id, "is_active": True})
        if not pay_plan:
            logger.warning(f"Can't find pay_plan with id: {_id}")
            return
        pay_plan = repo.PayPlansRepository.view.model_validate(pay_plan)
        await CacheRepository.set(f"payplan:{_id}", pay_plan, DEFAULT_TTL)
        return pay_plan


async def create_payment(telegram_id, engine=engine_r, **state_data):
    """⭐"""
    payplan_id = int(state_data["payplan_id"]) # callback data str type
    async with run_session(engine, commit=True) as session:
        user = await repo.UsersRepository.find_one_or_none(session, telegram_id=telegram_id)
        if not user:
            return
        pay_plan = await repo.PayPlansRepository.find_one_or_none(session, **{"id": payplan_id})
        pay_plan = repo.PayPlansRepository.view.model_validate(pay_plan)
        payment_method = state_data["payment_method"]
        for pay_plan_price in pay_plan.prices:
            if pay_plan_price.payment_method == payment_method:
                payment = await repo.PaymentsRepository.add(
                    session, schemas.Payment(
                        payplan_id=pay_plan.id,
                        payment_method=pay_plan_price.payment_method,
                        price=pay_plan_price.price_short,
                        currency=pay_plan_price.currency_code,
                        title=pay_plan.name
                        )
                    )
                user.payments.append(payment)
                return payment


async def _update_payment_info(payment_info, engine=engine_w):
    """⭐"""
    payment_id = UUID(payment_info.invoice_payload)
    async with run_session(engine) as session:
        _ = await repo.PaymentsRepository.update(
            session, {"id": payment_id},
            telegram_payment_charge_id=payment_info.telegram_payment_charge_id,
            provider_payment_charge_id=payment_info.provider_payment_charge_id,
            paid=True
            )
    return payment_id


async def _update_user_payment(payment_id, engine=engine_r):
    """⭐"""
    async with run_session(engine, commit=True) as session:
        user_payment = await repo.UserPaymentsRepository.find_one_or_none(session, payment_id=payment_id)
        user_balance_filters = {"user_id": user_payment.user.id, "currency": user_payment.payment.currency}
        user_balance = await repo.UserBalancesRepository.find_one_or_none(session, **user_balance_filters)
        if user_balance is None:
            user_balance = await repo.UserBalancesRepository.add(
                session, schemas.UserBalance(**user_balance_filters)
                )
        user_balance.amount += user_payment.payment.price
        await repo.UsersRepository.update(
            session, {"id": user_payment.user.id}, payplan_id=user_payment.payment.payplan_id
            )
    return user_payment, user_balance_filters


async def _update_user_connection(user_payment, engine=engine_r, **user_balance_filters):
    """⭐"""
    async with run_session(engine, commit=True) as session:
        pay_plan = await repo.PayPlansRepository.find_one_or_none(session, id=user_payment.payment.payplan_id)
        user_balance = await repo.UserBalancesRepository.find_one_or_none(session, **user_balance_filters)
        pay_plan_view = repo.PayPlansRepository.view.model_validate(pay_plan)
        _ = await update_user_marzban(
            user_payment.user.marzban_username, pay_plan.days_duration,
            data_limit=pay_plan_view.data_limit,
            data_limit_reset_strategy=pay_plan_view.data_limit_reset_strategy
            )
        user_balance.amount -= user_payment.payment.price
    return True


async def _update_user_referral(user_payment, bot, engine=engine_r, **user_balance_filters):
    """⭐"""
    referral = user_payment.user.referral
    if not(referral and referral.telegram_id): # and len(user_payment.user.payments) == 1
        return True
    async with run_session(engine, commit=True) as session:
        user = await get_user(referral.telegram_id)
        user_balance_filters["user_id"] = user.id
        user_balance = await repo.UserBalancesRepository.find_one_or_none(
            session, **user_balance_filters
            )
        if user_balance is None:
            user_balance = await repo.UserBalancesRepository.add(
                session, schemas.UserBalance(**user_balance_filters)
                )
        payed_amount = user_payment.payment.price * settings.business.referral_percent
        user_balance.amount += payed_amount

        await repo.ReferralPaymentsRepository.add(
            session, schemas.ReferralPayment(
                user_id=user.id, referral_id=referral.id,
                amount=payed_amount, currency=user_payment.payment.currency
                )
            )

    create_task(
        broadcast(
            bot,
            settings.telegram.messages["user-alert-referral"].format(payed_amount=payed_amount),
            referral.telegram_id
            )
        )

    return True


async def update_user_after_pay(user_id, payment_info, bot):
    """⭐"""
    await clean_cache(f"user:{user_id}")
    logger.info(f"Update payment info: {payment_info}")
    payment_id = await _update_payment_info(payment_info)
    logger.info(f"Update user balance for payment id: {payment_id}")
    user_payment, user_balance_filters = await _update_user_payment(payment_id)
    logger.info(f"Update user connection with id: {user_payment.user.id}")
    is_connection_updated = await _update_user_connection(user_payment, **user_balance_filters)
    if not is_connection_updated:
        logger.warning(f"Connection for user: {user_payment.user.id} not updated")
    try:
        is_referral_updated = await _update_user_referral(user_payment, bot, **user_balance_filters)
        if not is_referral_updated:
            logger.warning(f"Referral for user: {user_payment.user.id} not updated")
    except Exception as exc:
        logger.error(f"Can't update referral: {exc}")

    return user_payment


async def update_user_after_pay_no_provider(user_id, payment_id, engine=engine_r):
    """⭐"""
    await clean_cache(f"user:{user_id}")
    async with run_session(engine, commit=True) as session:
        _ = await repo.PaymentsRepository.update(session, {"id": payment_id}, paid=True)

        user_payment = await repo.UserPaymentsRepository.find_one_or_none(session, payment_id=payment_id)
        if user_payment is None:
            logger.warning(f"Can't find payment with id: {payment_id}")
            return

        payment_data = {"user_id": user_payment.user.id, "currency": user_payment.payment.currency}
        user_balance = await repo.UserBalancesRepository.find_one_or_none(session, **payment_data)
        _ = await repo.UsersRepository.update(
            session, {"id": user_payment.user.id}, payplan_id=user_payment.payment.payplan_id
            )

        await session.commit()

        pay_plan = await repo.PayPlansRepository.find_one_or_none(session, id=user_payment.payment.payplan_id)
        pay_plan_view = repo.PayPlansRepository.view.model_validate(pay_plan)
        _ = await update_user_marzban(
            user_payment.user.marzban_username, pay_plan.days_duration,
            data_limit=pay_plan_view.data_limit,
            data_limit_reset_strategy=pay_plan_view.data_limit_reset_strategy
            )
        user_balance.amount -= user_payment.payment.price

        await session.commit()

    return user_payment


async def get_invited_users(telegram_id, engine=engine_r):
    """⭐"""
    cache_result = await CacheRepository.get(f"invited:{telegram_id}")
    if cache_result:
        return cache_result
    async with run_session(engine) as session:
        users_count = await repo.ReferralRepository.get_stats_user(session, telegram_id)
        await CacheRepository.set(f"invited:{telegram_id}", users_count, DEFAULT_TTL)
        return users_count


async def get_coupon_and_coupon_usage(engine=engine_r, **filters):
    """⭐"""
    async with run_session(engine) as session:
        coupon = await repo.CouponsRepository.find_one_or_none(
            session, id=filters["coupon_id"], is_active=True
            )
        coupon_usage = await repo.UsageCouponsRepository.find_one_or_none(session, **filters)
        coupon_breaked_limit = False
        if coupon and coupon.usage_limit is not None:
            used_n = await repo.UsageCouponsRepository.get_stats(session, coupon_id=coupon.id)
            if used_n >= coupon.usage_limit:
                coupon_breaked_limit = True
        return coupon, coupon_usage, coupon_breaked_limit


async def create_coupon_usage(engine=engine_w, **values):
    """⭐"""
    async with run_session(engine) as session:
        coupon_usage_created = await repo.UsageCouponsRepository.add(
            session, repo.UsageCouponsRepository.schema(**values)
            )
        return coupon_usage_created


async def check_services_access(engine=engine_r) -> bool:
    """⭐"""
    async with ClientSession() as session, engine.connect() as connection:
        headers = await Marzban.auth(session)
        marzban_response = await Marzban.request(session, "GET", "/system", **headers)
        db_response = (await connection.execute(text("SELECT 1"))).scalar_one_or_none()
        cache_response = await CacheRepository.check_health()
        if marzban_response is not None and db_response is not None and cache_response:
            return True
    return False


async def get_user_marzban_status(session, username, **headers):
    """⭐"""
    result = await Marzban.request(session, "GET", f"/user/{username}", **headers)
    if result:
        return result["status"]


async def get_stats(engine=engine_r):
    """⭐"""
    cache_result = await CacheRepository.get("admin_stats")
    if cache_result:
        return cache_result
    async with \
        run_session(engine) as session, \
        ClientSession() as http_session, \
        run_session(engine_mrzbn) as session_marzban \
        :
        users_stats = {}
        users_total = await repo.UsersRepository.count(session)
        users_stats["users_total"] = users_total
        users_not_accept = await repo.UsersRepository.get_stats(session)
        users_stats["users_not_accept"] = users_not_accept

        marzban_names = await repo.UsersRepository.get_marzban_names(session)

        headers = await Marzban.auth(http_session)

        status_counters = []
        for sliced_marzban_names in slice_generator(marzban_names, 1000):
            marzban_statuses = await gather(
                *(
                    get_user_marzban_status(
                        http_session, username, **headers
                        ) for username in sliced_marzban_names
                    ), return_exceptions=True
                )
            status_counters.append(Counter(marzban_statuses))
            await asleep(1)
        if status_counters:
            status_counter = reduce(lambda x, y: x + y, status_counters)
        else:
            status_counter = {}
        users_stats["users_online"] = status_counter.get("active", 0)
        users_stats["users_offline"] = sum(
            status_counter.get(status, 0) for status in ("disabled", "expired", "on_hold")
            )

        users_table = await load_external_table(engine_mrzbn, "users")
        soon_expired_marzban_users = get_soon_expired_users(
            session_marzban, users_table, *[1, 2, 3]
            )
        soon_expired_telegram_users = get_telegram_expired_users(session, soon_expired_marzban_users)
        async for days_exp, telegram_ids in soon_expired_telegram_users:
            users_stats[f"users_expired_{days_exp}d"] = len(telegram_ids)

        pay_stats = await repo.PaymentsRepository.get_stats(session)

        coupons_usage_data = await repo.UsageCouponsRepository.get_stats_grouped(session)
        coupons_usage = []
        for c1, c2 in coupons_usage_data:
            coupons_usage.append(f"{c1}: {c2}")
        users_stats["coupons_usage"] = "\n".join(coupons_usage)
        stats = {**users_stats, **pay_stats}
        await CacheRepository.set("admin_stats", stats, DEFAULT_TTL)
        return stats


async def send_stats(bot, telegram_id):
    """⭐"""
    stats = await get_stats()
    await bot.send_message(
        telegram_id,
        settings.telegram.messages["admin-stats"].format(**stats)
        )


async def notify_users(bot, engine=engine_w, **state_data):
    """⭐"""
    notify_group = state_data["notify_group"]
    if notify_group == "all":
        async with run_session(engine) as session:
            users = await repo.UsersRepository.find_all(session)
            users = [user.telegram_id for user in users if user.marzban_username]
    elif notify_group in ("active", "expired"):
        async with \
            run_session(engine) as session, \
            run_session(engine_mrzbn) as session_marzban \
            :
            users_table = await load_external_table(engine_mrzbn, "users")
            marzban_names = await get_marzban_users(session_marzban, users_table, notify_group)
            users = await repo.UsersRepository.get_telegram_ids_by_marzban_names(
                session, marzban_names
                )
    elif notify_group in ("pay-expired", "pay-no-expired"):
        async with \
            run_session(engine) as session, \
            run_session(engine_mrzbn) as session_marzban \
            :
            users_table = await load_external_table(engine_mrzbn, "users")
            marzban_names = await get_marzban_users(session_marzban, users_table, "expired")
            users = []
            if notify_group == "pay-expired":
                users_generator = repo.UsersRepository.get_telegram_ids_by_marzban_names_exp
            else:
                users_generator = repo.UsersRepository.get_telegram_ids_by_marzban_names_exp_no_pay
            async for telegram_id in users_generator(session, marzban_names):
                users.append(telegram_id)
    else:
        users = []
    await broadcast(
        bot, state_data["notify_message"], *users, photo=state_data.get("notify_image")
        )
    return len(users)


async def create_coupon(engine=engine_w, **state_data):
    """⭐"""
    values = {
        "id": state_data["coupon_id"],
        "days_duration": int(state_data["coupon_days"])
        }
    usage_limit = state_data.get("coupon_usage_limit")
    if usage_limit:
        values["usage_limit"] = int(usage_limit)
    async with run_session(engine) as session:
        coupon_exists = await repo.CouponsRepository.find_one_or_none(
            session, **{"id": values["id"]}
            )
        if coupon_exists:
            return
        coupon_created = await repo.CouponsRepository.add(
            session, repo.CouponsRepository.schema(**values)
            )
        return coupon_created


async def get_referral_balance(user_id, engine=engine_r):
    """⭐"""
    cache_result = await CacheRepository.get(f"balance:{user_id}")
    if cache_result:
        return cache_result
    async with run_session(engine) as session:
        user_balance = await repo.ReferralPaymentsRepository.get_user_balance(session, user_id)
        user_balance = round(user_balance, 1) if user_balance else 0
        await CacheRepository.set(f"balance:{user_id}", user_balance, DEFAULT_TTL)
        return user_balance


async def check_user_balance(telegram_id, pay_plan_currency, pay_plan_price, engine=engine_r):
    """⭐"""
    async with run_session(engine) as session:
        user = await repo.UsersRepository.find_one_or_none(session, telegram_id=telegram_id)
        if not user:
            return False
        for balance in user.balances:
            if balance.currency == pay_plan_currency:
                if balance.amount >= pay_plan_price:
                    return True
    return False


async def clean_cache(key: str):
    """⭐"""
    cache_result = await CacheRepository.get(key)
    if cache_result:
        await CacheRepository.delete(key)


def define_app_by_url(url: str):
    """⭐"""
    for app in settings.business.supported_apps:
        if app in url:
            return settings.business.supported_apps[app]



# -*- coding: utf-8 -*-
from aiogram.filters import BaseFilter
from aiogram.types import Message

from vpnflow.settings import settings


class AdminFilter(BaseFilter):
    """⭐"""

    ADMINS = settings.telegram.admins_id

    async def __call__(self, event: Message) -> bool:
        user = event.from_user
        if user:
            return user.id in self.ADMINS
        return False


class BlacklistFilter(BaseFilter):
    """⭐"""

    BLACKLIST = set(settings.telegram.blacklist)

    async def __call__(self, event: Message) -> bool:
        user = event.from_user
        if user:
            return False if user.id in self.BLACKLIST else True



# -*- coding: utf-8 -*-
from datetime import datetime
from logging import getLogger

from aiogram.types import Update
from fastapi import APIRouter, Request
from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
from fastapi.responses import HTMLResponse, JSONResponse
from pydantic_core._pydantic_core import ValidationError

from vpnflow.settings import settings

logger = getLogger(__name__)

base, swagger, telegram = APIRouter(), APIRouter(), APIRouter(prefix="/api/webhook")
ROUTERS = [base, telegram]
if settings.webserver.swagger_routes:
    ROUTERS.append(swagger)


@base.get("/", description="Index page")
def index(request: Request):
    """⭐"""
    return {"ts": datetime.utcnow()}


@swagger.get("/docs", include_in_schema=False)
async def docs(request: Request) -> HTMLResponse:
    """⭐"""
    return get_swagger_ui_html(
        openapi_url=request.app.openapi_url, title=request.app.title,
        swagger_js_url=f"{settings.webserver.static_url}/js/swagger-ui-bundle.js",
        swagger_css_url=f"{settings.webserver.static_url}/css/swagger-ui.css",
        swagger_favicon_url=f"{settings.webserver.static_url}/icons/favicon.ico",
        )


@swagger.get("/redoc", include_in_schema=False)
async def redoc(request: Request) -> HTMLResponse:
    """⭐"""
    return get_redoc_html(
        openapi_url=request.app.openapi_url, title=request.app.title, with_google_fonts=False,
        redoc_js_url=f"{settings.webserver.static_url}/js/swagger-redoc.standalone.js",
        redoc_favicon_url=f"{settings.webserver.static_url}/icons/favicon.ico"
        )


@telegram.post("/")
async def bot_webhook(request: Request):
    """⭐"""
    resp_client_400 = JSONResponse({"message": "Bad request"}, status_code=400)
    logger.debug("Webhook processing start")
    rq_token = request.headers.get("X-Telegram-Bot-Api-Secret-Token")
    if rq_token != settings.telegram.webhook_secret_token.get_secret_value():
        logger.error(f"Bad request from {request.client.host}")
        return resp_client_400
    update = await request.json()
    try:
        update = Update.model_validate(update, context={"bot": request.app.bot})
    except ValidationError as exc:
        logger.exception(exc)
        logger.error(f"Bad request from {request.client.host}")
        return resp_client_400
    else:
        await request.app.bot_dispatcher.feed_update(request.app.bot, update)
        logger.debug("Webhook processing end")
        return JSONResponse({"message": "ok"}, status_code=200)



# -*- coding: utf-8 -*-
from logging import getLogger

from sqlalchemy import event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

logger = getLogger(__name__)


@event.listens_for(Engine, "connect")
def on_connect(conn, conn_record):
    """⭐"""
    logger.debug(f"Open DBAPI connection: {conn}")


@event.listens_for(Engine, "close")
def on_close(conn, conn_record):
    """⭐"""
    logger.debug(f"Close DBAPI connection: {conn}")


@event.listens_for(Engine, "before_execute")
def on_before_execute(
    conn,
    clauseelement,
    multiparams,
    params,
    execution_options
    ):
    """⭐"""
    logger.debug("Before execute")


@event.listens_for(Session, "before_commit")
def on_before_commit(session):
    """⭐"""
    logger.debug("Before commit")


@event.listens_for(Session, "after_commit")
def on_after_commit(session):
    """⭐"""
    logger.debug("Before commit")



# -*- coding: utf-8 -*-
from datetime import datetime
from decimal import Decimal
from typing import List, Optional
from uuid import UUID, uuid4

from sqlalchemy import (DECIMAL, TIMESTAMP, VARCHAR, BigInteger, Boolean,
                        ForeignKey, Index, Integer, Text, UniqueConstraint,
                        Uuid, func)
from sqlalchemy.orm import Mapped, mapped_column, relationship

from vpnflow.db.base import Base


class User(Base):
    """⭐"""

    __tablename__ = "users"
    __table_args__ = (UniqueConstraint("referral_id"),)

    id: Mapped[int] = mapped_column(Integer, unique=True, primary_key=True, autoincrement=True)
    ts_created: Mapped[datetime] = mapped_column(TIMESTAMP, server_default=func.now())
    ts_updated: Mapped[datetime] = mapped_column(TIMESTAMP, server_default=func.now(), onupdate=func.now())
    telegram_id: Mapped[int] = mapped_column(ForeignKey("users_telegram.id"))
    marzban_username: Mapped[Optional[str]] = mapped_column(VARCHAR(64))
    payplan_id: Mapped[Optional[int]] = mapped_column(Integer, nullable=True)
    referral_id: Mapped[Optional[int]] = mapped_column(ForeignKey("referrals.id"))

    user_telegram: Mapped["UserTelegram"] = relationship(back_populates="user")
    referral: Mapped["Referral"] = relationship(back_populates="user", lazy="selectin")
    usages_coupons_association: Mapped[List["UsageCoupon"]] = relationship(back_populates="user")
    coupons: Mapped[List["Coupon"]] = relationship(secondary="usage_coupons", back_populates="users")
    user_payments_association: Mapped[List["UserPayment"]] = relationship(back_populates="user")
    payments: Mapped[List["Payment"]] = relationship(secondary="user_payments", back_populates="users", lazy="selectin")
    balances: Mapped[List["UserBalance"]] = relationship(back_populates="user", lazy="selectin")

    referral_payment_association: Mapped[List["ReferralPayment"]] = relationship(back_populates="user")
    referrals: Mapped[List["Referral"]] = relationship(secondary="referral_payments", back_populates="users")


class UserTelegram(Base):
    """⭐"""

    __tablename__ = "users_telegram"

    id: Mapped[int] = mapped_column(BigInteger, unique=True, primary_key=True)
    name: Mapped[Optional[str]] = mapped_column(VARCHAR(64))
    name_full: Mapped[Optional[str]] = mapped_column(VARCHAR(64))
    language_code: Mapped[Optional[str]] = mapped_column(VARCHAR(16))
    is_bot: Mapped[Optional[bool]] = mapped_column(Boolean)
    is_premium: Mapped[Optional[bool]] = mapped_column(Boolean)

    user: Mapped["User"] = relationship(back_populates="user_telegram")


class Referral(Base):
    """⭐"""

    __tablename__ = "referrals"

    id: Mapped[int] = mapped_column(Integer, unique=True, primary_key=True, autoincrement=True)
    telegram_start: Mapped[str] = mapped_column(Text)
    telegram_id: Mapped[Optional[int]] = mapped_column(BigInteger)

    user: Mapped["User"] = relationship(back_populates="referral")
    referral_payment_association: Mapped[List["ReferralPayment"]] = relationship(back_populates="referral")
    users: Mapped[List["User"]] = relationship(secondary="referral_payments", back_populates="referrals")


class PayPlan(Base):
    """⭐"""

    __tablename__ = "pay_plans"

    id: Mapped[int] = mapped_column(Integer, unique=True, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(VARCHAR(64), nullable=False)
    days_duration: Mapped[int] = mapped_column(Integer, nullable=False)
    data_limit_gb: Mapped[int] = mapped_column(Integer, nullable=False)
    data_limit_reset_strategy: Mapped[str] = mapped_column(VARCHAR(16), nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, default=True)

    prices: Mapped[List["PayPlanPrice"]] = relationship(back_populates="pay_plan", lazy="selectin")


class PayPlanPrice(Base):
    """⭐"""

    __tablename__ = "payplan_prices"

    payplan_id: Mapped[int] = mapped_column(ForeignKey("pay_plans.id"), primary_key=True)
    payment_method: Mapped[str] = mapped_column(VARCHAR(32), primary_key=True)
    price: Mapped[Decimal] = mapped_column(DECIMAL)
    currency_code: Mapped[str] = mapped_column(ForeignKey("price_currencies.code"), primary_key=True)

    pay_plan: Mapped["PayPlan"] = relationship(back_populates="prices")
    currency: Mapped["PayPlanPriceCurrency"] = relationship(back_populates="price", lazy="selectin")


class PayPlanPriceCurrency(Base):
    """⭐"""

    __tablename__ = "price_currencies"

    code: Mapped[str] = mapped_column(VARCHAR(16), primary_key=True, unique=True)
    symbol: Mapped[str] = mapped_column(VARCHAR(8))

    price: Mapped["PayPlanPrice"] = relationship(back_populates="currency")


class Coupon(Base):
    """⭐"""

    __tablename__ = "coupons"

    id: Mapped[str] = mapped_column(VARCHAR(24), unique=True, primary_key=True)
    days_duration: Mapped[int] = mapped_column(Integer, nullable=False)
    is_active: Mapped[bool] = mapped_column(Boolean, default=True)
    usage_limit: Mapped[Optional[int]] = mapped_column(Integer, nullable=True)

    usages_coupons_association: Mapped[List["UsageCoupon"]] = relationship(back_populates="coupon")
    users: Mapped[List["User"]] = relationship(secondary="usage_coupons", back_populates="coupons")


class UsageCoupon(Base):
    """⭐"""

    __tablename__ = "usage_coupons"

    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), primary_key=True)
    coupon_id: Mapped[str] = mapped_column(ForeignKey("coupons.id"), primary_key=True)

    user: Mapped["User"] = relationship(back_populates="usages_coupons_association")
    coupon: Mapped["Coupon"] = relationship(back_populates="usages_coupons_association")


class Payment(Base):
    """⭐"""

    __tablename__ = "payments"

    id: Mapped[UUID] = mapped_column(Uuid, unique=True, primary_key=True, default=uuid4)
    ts_created: Mapped[datetime] = mapped_column(TIMESTAMP, server_default=func.now())
    ts_updated: Mapped[datetime] = mapped_column(TIMESTAMP, server_default=func.now(), onupdate=func.now())
    payplan_id: Mapped[int] = mapped_column(Integer, nullable=False)
    payment_method: Mapped[str] = mapped_column(VARCHAR(32), nullable=False)
    price: Mapped[int] = mapped_column(Integer, nullable=False)
    currency: Mapped[str] = mapped_column(VARCHAR(16), nullable=False)
    paid: Mapped[bool] = mapped_column(Boolean, default=False)
    title: Mapped[str] = mapped_column(VARCHAR(128), nullable=False)
    telegram_payment_charge_id: Mapped[Optional[str]] = mapped_column(VARCHAR(64), nullable=True)
    provider_payment_charge_id: Mapped[Optional[str]] = mapped_column(VARCHAR(64), nullable=True)

    user_payments_association: Mapped[List["UserPayment"]] = relationship(back_populates="payment")
    users: Mapped[List["User"]] = relationship(secondary="user_payments", back_populates="payments")


class UserPayment(Base):
    """⭐"""

    __tablename__ = "user_payments"

    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), primary_key=True)
    payment_id: Mapped[UUID] = mapped_column(ForeignKey("payments.id"), primary_key=True)

    user: Mapped["User"] = relationship(back_populates="user_payments_association", lazy="selectin")
    payment: Mapped["Payment"] = relationship(back_populates="user_payments_association", lazy="selectin")


class UserBalance(Base):
    """⭐"""

    __tablename__ = "users_balances"

    id: Mapped[int] = mapped_column(Integer, unique=True, primary_key=True, autoincrement=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"))
    currency: Mapped[str] = mapped_column(VARCHAR(16), nullable=False)
    amount: Mapped[Decimal] = mapped_column(DECIMAL, default=Decimal(0))

    user: Mapped["User"] = relationship(back_populates="balances")


class ReferralPayment(Base):
    """⭐"""

    __tablename__ = "referral_payments"

    ts_created: Mapped[datetime] = mapped_column(TIMESTAMP, server_default=func.now())
    user_id: Mapped[int] = mapped_column(ForeignKey("users.id"), primary_key=True)
    referral_id: Mapped[int] = mapped_column(ForeignKey("referrals.id"), primary_key=True)
    amount: Mapped[Decimal] = mapped_column(DECIMAL, nullable=False)
    currency: Mapped[str] = mapped_column(VARCHAR(16), nullable=False)

    user: Mapped["User"] = relationship(back_populates="referral_payment_association")
    referral: Mapped["Referral"] = relationship(back_populates="referral_payment_association")


indexes = (
    Index("idx_users_telegram_id", User.telegram_id),
)



# -*- coding: utf-8 -*-
from asyncio import get_event_loop
from logging import getLogger

import uvicorn
from aiogram.exceptions import TelegramRetryAfter

from vpnflow.bot.base import create_dispatcher
from vpnflow.cli import create_parser, show_art
from vpnflow.db.tools import init_db
from vpnflow.flows._schedule import run_scheduled_tasks
from vpnflow.misc import load_log_conf
from vpnflow.settings import settings

logger = getLogger(__name__)


def main():
    """⭐"""
    show_art()
    parser, loop = create_parser(), get_event_loop()

    args = parser.parse_args()

    if args.log_conf_file:
        load_log_conf(args.log_conf_file)

    logger.debug(f"Run with args: {args}, settings: {settings}")

    if args.run_scheduled_tasks:
        run_scheduled_tasks()

    if args.command == "db":
        loop.run_until_complete(init_db(args))

    if settings.telegram.webhook_use:
        uvicorn.run(
            "vpnflow.web.app:app",
            host=settings.webserver.host, port=settings.webserver.port,
            workers=settings.webserver.workers, reload=settings.webserver.reload
            )
    else:
        bot_dispatcher = create_dispatcher()
        try:
            loop.run_until_complete(
                bot_dispatcher.start_polling(
                    bot_dispatcher.bot,
                    allowed_updates=bot_dispatcher.resolve_used_update_types(),
                    polling_timeout=60, skip_updates=True
                    )
                )
        except KeyboardInterrupt:
            loop.run_until_complete(bot_dispatcher.stop_polling())
        except TelegramRetryAfter as exc:
            logger.error(exc)
        except Exception as exc:
            logger.exception(exc)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from aiogram.fsm.state import State, StatesGroup


class CouponForm(StatesGroup):
    """⭐"""
    coupon = State()


class PayPlanForm(StatesGroup):
    """⭐"""
    payplan_choice = State()
    pay_method_choice = State()
    accept_choice = State()
    write_off_balance = State()


class SupportForm(StatesGroup):
    """⭐"""
    question = State()


class NotifyForm(StatesGroup):
    """⭐"""
    group_choice = State()
    message = State()
    accept_choice = State()


class CouponAddForm(StatesGroup):
    """⭐"""
    name = State()
    days = State()
    accept_choice = State()
    accept_usage_limit = State()
    usage_limit = State()



# -*- coding: utf-8 -*-
from logging import getLogger
from typing import Optional, Tuple

from aiogram import Bot, Dispatcher
from aiogram.client.default import DefaultBotProperties
from aiogram.client.session.aiohttp import AiohttpSession
from aiogram.enums import ParseMode
from aiogram.fsm.storage.base import BaseStorage
from aiogram.fsm.storage.memory import MemoryStorage
from aiogram.fsm.storage.redis import RedisStorage
from aiogram.types.bot_command_scope_chat import BotCommandScopeChat
from aiogram.utils.callback_answer import CallbackAnswerMiddleware

from vpnflow.bot import commands, middleware, routers
from vpnflow.services import _redis
from vpnflow.services._telegram import broadcast
from vpnflow.services.cache import CacheRepository
from vpnflow.settings import Settings, TelegramSettings, settings

logger = getLogger(__name__)
telegram_settings = settings.telegram


async def on_startup_bot(
    bot: Bot,
    bot_dispatcher: Dispatcher,
    settings_telegram: TelegramSettings = telegram_settings
    ):
    """⭐"""
    logger.info("Startup bot")
    bot_info = await bot.get_me()
    logger.info(f"Name: {bot_info.full_name}. Username: @{bot_info.username}. ID: {bot_info.id}")
    assert bot_info.username == settings.telegram.bot_name
    await broadcast(bot, settings_telegram.messages["bot-on"], *settings_telegram.admins_id)
    await bot.delete_my_commands()
    await bot.set_my_commands(commands.user)
    for admin_id in settings_telegram.admins_id:
        await bot.set_my_commands(commands.admin, scope=BotCommandScopeChat(chat_id=admin_id))
    if settings_telegram.webhook_use:
        webhook_info = await bot.get_webhook_info()
        logger.info(f"Webhook info: {webhook_info}")
        if webhook_info.url != settings_telegram.webhook_url:
            await bot.delete_webhook(drop_pending_updates=True)
            await bot.set_webhook(
                url=settings_telegram.webhook_url, drop_pending_updates=True,
                allowed_updates=bot_dispatcher.resolve_used_update_types(),
                max_connections=settings_telegram.webhook_max_connections,
                secret_token=settings_telegram.webhook_secret_token.get_secret_value()
                )
    is_health = await CacheRepository.check_health()
    assert is_health


async def on_shutdown_bot(
    bot: Bot,
    bot_dispatcher: Dispatcher,
    settings_telegram: TelegramSettings = telegram_settings
    ):
    """⭐"""
    logger.info("Shutdown bot")
    if isinstance(bot_dispatcher.storage, RedisStorage):
        await bot_dispatcher.storage.close()
        await bot_dispatcher.fsm.storage.close()
    await broadcast(bot, settings_telegram.messages["bot-off"], *settings_telegram.admins_id)
    # if settings_telegram.webhook_use:
    #     await bot.delete_webhook(drop_pending_updates=True)
    await bot.session.close()
    await bot.close()
    await CacheRepository.close()


async def on_startup_bot_dispatcher(dispatcher) -> None:
    """⭐"""
    logger.info("Startup bot dispatcher")
    await on_startup_bot(dispatcher.bot, dispatcher)


async def on_shutdown_bot_dispatcher(dispatcher) -> None:
    """⭐"""
    logger.info("Shutdown bot dispatcher")
    await on_shutdown_bot(dispatcher.bot, dispatcher)


def create_bot(
    settings_telegram: TelegramSettings = telegram_settings,
    session: Optional[AiohttpSession] = None
    ) -> Bot:
    """⭐"""
    if session is None:
        session: AiohttpSession = AiohttpSession()
    return Bot(
        token=settings_telegram.bot_token.get_secret_value(),
        default=DefaultBotProperties(parse_mode=ParseMode.HTML),
        session=session
        )


def create_dispatcher(
    bot: Bot = create_bot(),
    middlewares: Tuple = middleware.ALL,
    routers: Tuple = routers.ALL,
    settings: Settings = settings,
    on_startup = on_startup_bot_dispatcher,
    on_shutdown = on_shutdown_bot_dispatcher
    ) -> Dispatcher:
    """⭐"""
    redis_url = settings.telegram.redis_url.get_secret_value()
    if redis_url:
        storage: BaseStorage = RedisStorage(redis=_redis.create_client(redis_url))
    else:
        storage: BaseStorage = MemoryStorage()
    dispatcher = Dispatcher(name="bot_dispatcher", storage=storage)
    dispatcher.bot = bot
    for _middleware in middlewares:
        dispatcher.update.middleware.register(_middleware)
    dispatcher.callback_query.middleware(CallbackAnswerMiddleware())
    dispatcher.include_routers(*routers)
    dispatcher.startup.register(on_startup)
    dispatcher.shutdown.register(on_shutdown)
    return dispatcher



# -*- coding: utf-8 -*-
import random
from logging import getLogger
from ssl import create_default_context

import certifi
from aiohttp import FormData

from vpnflow.settings import settings

logger = getLogger(__name__)
ssl_context = create_default_context(cafile=certifi.where())


class MarzbanUsernameGenerator:
    """⭐"""

    def __init__(self, name_len: int = settings.business.username_len):
        self._len = name_len
        self._alphabet = ("aeiou", "bcdfghjklmnpqrstvwxyz") # vowels, consonants

    def __next__(self):
        word = ""
        for _ in range(self._len):
            if _ % 2 == 0:
                word += random.choice(self._alphabet[1])
            else:
                word += random.choice(self._alphabet[0])
        return word.upper()

    def __call__(self):
        return next(self)


class Marzban:
    """⭐"""

    settings = settings.marzban
    usernames_generator = MarzbanUsernameGenerator()

    @classmethod
    async def auth(cls, session):
        """⭐"""
        auth_data = FormData()
        auth_data.add_field('username', cls.settings.username)
        auth_data.add_field('password', cls.settings.password.get_secret_value())
        rs = await cls.request(session, "POST", "/admin/token", json=None, data=auth_data)
        return {"Authorization": "Bearer {access_token}".format(**rs)}

    @classmethod
    async def request(cls, session, method, path, json=None, data=None, **headers):
        """⭐"""
        if json:
            session_request = session.request(
                method, cls.settings.host_api + path,
                json=json, headers=headers, ssl=ssl_context
                )
        elif data:
            session_request = session.request(
                method, cls.settings.host_api + path,
                data=data, headers=headers, ssl=ssl_context
                )
        else:
            session_request = session.request(
                method, cls.settings.host_api + path,
                headers=headers, ssl=ssl_context
                )
        async with session_request as resp:
            if resp.status == 200:
                resp_data = await resp.json()
                return resp_data
            else:
                try:
                    resp_data = await resp.json()
                except Exception:
                    resp_data = await resp.text()
                logger.error(f"Status: {resp.status}. Message: {resp_data}")
                resp.raise_for_status()



# -*- coding: utf-8 -*-
from datetime import timezone
from decimal import Decimal
from typing import Dict, List, Optional, final

from pydantic import SecretStr, ValidationInfo, computed_field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from vpnflow import __author__, __email__, __version__
from vpnflow.misc import load_yaml


class SecretMixin:
    """⭐"""

    class Config:
        """⭐"""

        json_encoders = {SecretStr: lambda v: v.get_secret_value() if v else None}
        extra = "ignore"


class _BaseSettings(BaseSettings, SecretMixin):
    """⭐"""


class DatabaseSettings(_BaseSettings):
    """⭐"""
    sqlalchemy_url: SecretStr
    sqlalchemy_engine_params: Dict = {}


class CacheSettings(_BaseSettings):
    """⭐"""
    redis_url: Optional[SecretStr] = ""
    default_ttl: int = 10


class WebServerSettings(_BaseSettings):
    """⭐"""
    host: str = "127.0.0.1"
    port: int = 80
    workers: int = 1
    reload: bool = False
    debug: bool = False
    docs_url: str = ""
    redoc_url: str = ""
    openapi_url: str = "/openapi.json"
    version: str = __version__
    title: str = "vpnflow"
    description: str = ""
    summary: str = ""
    contact: dict = {"name": __author__, "email": __email__}
    static_url: str = ""
    static_dir: str = ""
    templates_dir: str = ""
    swagger_routes: bool = False


class TelegramSettings(_BaseSettings):
    """⭐"""
    bot_name: str
    bot_token: SecretStr
    admins_id: List[int] = []
    log_chat: Optional[int] = None
    provider_token_yookassa: SecretStr
    provider_need_email: bool = True
    provider_send_email_to_provider: bool = True
    rate_limit: float = 0.5
    webhook_use: Optional[bool] = False
    webhook_url_base: str = ""
    webhook_path: str = ""
    webhook_port: int = -1
    webhook_url: str = ""
    webhook_max_connections: int = 50
    webhook_secret_token: Optional[SecretStr] = ""
    messages_path: str = ""
    messages: Dict = {}
    redis_url: Optional[SecretStr] = ""
    blacklist: List = []

    @field_validator("bot_token", mode="before")
    def validate_bot_token(cls, raw: str) -> str:
        """⭐"""
        assert len(raw) == 46
        s1, s2 = raw.split(":")
        assert len(s1) == 10 and len(s2) == 35
        return raw

    @field_validator("webhook_url", mode="before")
    def build_webhook_url(cls, value, info: ValidationInfo) -> str:
        """⭐"""
        values = info.data
        if values.get("webhook_use"):
            url_base = values['webhook_url_base']
            url_path = values['webhook_path']
            port = values['webhook_port']
            if "localhost" in url_base:
                return f"{url_base}:{port}{url_path}"
            return f"{url_base}{url_path}"
        return ""

    @field_validator("messages", mode="before")
    def load_messages(cls, value, info: ValidationInfo) -> dict:
        """⭐"""
        values = info.data
        if values.get("messages_path"):
            return load_yaml(values.get("messages_path"))
        return {}


class MarzbanSettings(_BaseSettings):
    """⭐"""
    host_api: str
    username: str
    password: SecretStr
    token_expire: int
    verify_ssl: bool
    sqlalchemy_url: Optional[SecretStr] = ""


class BusinessSettings(_BaseSettings):
    """⭐"""
    referral_percent: Decimal = Decimal(0)
    test_days: int
    test_traffic_gb_limit: int
    test_proxies: Dict = {}
    test_inbounds: Dict
    url_user_agreement: str
    url_privacy_policy: str
    url_faq: str
    url_support: str
    url_redirect: str
    primary_currency: str = "RUB"
    supported_platforms: Dict
    supported_apps: Dict
    default_desc: Optional[str] = ""
    username_len: Optional[int] = 10

    @computed_field
    def referral_percent_text(self) -> str:
        """⭐"""
        return str(int(self.referral_percent * 100))


class FlowsSettings(_BaseSettings):
    """⭐"""
    notify_users_cron: str = ""
    notify_users_days: List[int] = []
    notify_users_retry_max: int = 1
    notify_users_retry_delay: int = 1


@final
class Settings(BaseSettings, SecretMixin):
    """⭐"""
    tz: timezone = timezone.utc
    database: DatabaseSettings
    cache: CacheSettings
    webserver: WebServerSettings = {}
    telegram: TelegramSettings
    marzban: MarzbanSettings = {}
    business: BusinessSettings
    flows: FlowsSettings = {}

    model_config = SettingsConfigDict(
        extra="allow", env_file=".env",
        env_nested_delimiter="__", env_file_encoding="utf-8",
    ) # env_prefix="APP_"


settings = Settings()



# -*- coding: utf-8 -*-
import datetime
from logging import getLogger

from sqlalchemy import MetaData, Table, and_, select

from vpnflow.db import repositories as repo
from vpnflow.db.base import Base, engine_w, run_session
from vpnflow.db.models import User, indexes
from vpnflow.misc import load_yaml

logger = getLogger(__name__)


async def load_external_table(engine, table_name):
    """⭐"""
    async with engine.begin() as conn:
        table = await conn.run_sync(
            lambda conn: Table(table_name, MetaData(), autoload_with=conn)
            )
        logger.debug(f"Table params: {table.params}\nTable keys: {table.c.keys()}")
    return table


async def init_db(args, engine=engine_w):
    """⭐"""
    async with engine.begin() as conn:
        if args.drop:
            await conn.run_sync(Base.metadata.drop_all, checkfirst=True)
        if args.create:
            await conn.run_sync(Base.metadata.create_all, checkfirst=True)
            for index in indexes:
                await conn.run_sync(index.create, checkfirst=True)
    if args.insert:
        async with run_session(engine) as session:
            insert_data = load_yaml(args.insert)

            _ = await repo.PayPlanPriceCurrenciesRepository.add_many(
                session,
                [
                    repo.PayPlanPriceCurrenciesRepository.schema(
                        **row
                        ) for row in insert_data.get("price_currencies", [])
                    ]
                )
            for row in insert_data.get("pay_plans_and_prices", []):
                pay_plan = await repo.PayPlansRepository.add(session, repo.PayPlansRepository.schema(**row))
                _ = await repo.PayPlanPricesRepository.add_many(
                    session,
                    [
                        repo.PayPlanPricesRepository.schema(
                            **{**row_p, **{"payplan_id": pay_plan.id}}
                            ) for row_p in row.get("prices", [])
                        ]
                    )
            _ = await repo.CouponsRepository.add_many(
                session,
                [
                    repo.CouponsRepository.schema(
                        **coupon
                        ) for coupon in insert_data.get("coupons", [])
                    ]
                )


async def get_soon_expired_users(session, users_table, *args):
    """⭐"""
    dt = datetime.datetime.utcnow()
    dt_up = (
        datetime
        .datetime(dt.year, dt.month, dt.day, 23, 59)
        .replace(tzinfo=datetime.timezone.utc)
    )
    for days_future in args:
        _dt_up = (dt_up + datetime.timedelta(days=days_future)).timestamp()
        users = await session.execute(
            select(
                # users_table.c.id,
                users_table.c.username
            )
            .where(
                and_(
                    users_table.c.status == "active",
                    users_table.c.expire >= _dt_up - 86400,
                    users_table.c.expire <= _dt_up
                    )
                )
            )
        yield days_future, (row.username for row in users.fetchall())


async def get_telegram_expired_users(session, soon_expired_users):
    """⭐"""
    async for days_exp, usernames in soon_expired_users:
        query = select(User.telegram_id).where(User.marzban_username.in_(usernames))
        users_id = (await session.execute(query)).scalars().all()
        yield days_exp, users_id


async def get_marzban_users(session, users_table, status):
    """⭐"""
    users = await session.execute(
        select(users_table.c.username).where(users_table.c.status == status)
        )
    return (row.username for row in users.fetchall())



# -*- coding: utf-8 -*-
from datetime import datetime, timedelta
from logging import getLogger
from typing import Generic, List, Optional, Sequence, Type, TypeVar

from sqlalchemy import delete, func, select, update
from sqlalchemy.ext.asyncio import AsyncSession

from vpnflow.db import models as orm
from vpnflow.db.base import Base
from vpnflow.settings import settings
from vpnflow.web import schemas

logger = getLogger(__name__)

ModelORM = TypeVar("ModelORM", Base, Base)
ModelSchema = TypeVar("ModelSchema")


class Repository(Generic[ModelORM, ModelSchema]):
    """⭐"""

    model: Type[ModelORM]
    schema: Type[ModelSchema]

    @classmethod
    async def add(cls, session: AsyncSession, instance: ModelSchema) -> ModelORM:
        """⭐"""
        values = instance.model_dump(exclude_unset=True)
        model_obj = cls.model(**values)
        session.add(model_obj)
        await session.flush()
        # await session.commit()
        await session.refresh(model_obj)
        return model_obj

    @classmethod
    async def add_many(cls, session: AsyncSession, instances: List[ModelSchema]):
        """⭐"""
        model_objs = (
            cls.model(
                **instance.model_dump(exclude_unset=True)
                ) for instance in instances
            )
        session.add_all(model_objs)
        await session.flush()
        # await session.commit()
        return model_objs

    @classmethod
    async def find_one_or_none(cls, session: AsyncSession, **filter_by) -> Optional[ModelORM]:
        """⭐"""
        query = select(cls.model).filter_by(**filter_by)
        logger.debug(query)
        result = await session.execute(query)
        return result.scalar_one_or_none()

    @classmethod
    async def find_all(cls, session: AsyncSession, **filter_by) -> Sequence[ModelORM]:
        """⭐"""
        query = select(cls.model).filter_by(**filter_by)
        logger.debug(query)
        result = await session.execute(query)
        return result.scalars().all()

    @classmethod
    async def update(cls, session: AsyncSession, filter_by, **values):
        """⭐"""
        query = (
            update(cls.model)
            .where(*[getattr(cls.model, k) == v for k, v in filter_by.items()])
            .values(**values)
            .execution_options(synchronize_session="fetch")
            # .returning(cls.model)
        )
        logger.debug(query)
        result = await session.execute(query)
        # await session.commit()
        logger.debug(f"Updated {result.rowcount} rows")
        return result

    @classmethod
    async def delete(cls, session: AsyncSession, **filter_by):
        """⭐"""
        query = delete(cls.model).filter_by(**filter_by)
        logger.debug(query)
        result = await session.execute(query)
        # await session.commit()
        logger.debug(f"Deleted {result.rowcount} rows")
        return result

    @classmethod
    async def delete_all(cls, session: AsyncSession):
        """⭐"""
        result = await session.execute(delete(cls.model))
        return result

    @classmethod
    async def count(cls, session: AsyncSession):
        """⭐"""
        result = await session.execute(
            select(func.count()).select_from(cls.model)
            )
        return result.scalar_one_or_none()

    @staticmethod
    async def commit(session: AsyncSession) -> None:
        """⭐"""
        await session.commit()


class UsersTelegramRepository(Repository):
    """⭐"""

    model, schema = orm.UserTelegram, schemas.UserTelegram


class ReferralRepository(Repository):
    """⭐"""

    model, schema = orm.Referral, schemas.Referral

    @classmethod
    async def get_stats_user(cls, session: AsyncSession, telegram_id: int):
        """⭐"""
        query = (
            select(func.count("*").label("c"))
            .where(cls.model.telegram_id == telegram_id)
            )
        result = await session.execute(query)
        return result.scalar_one_or_none()


class UsersRepository(Repository):
    """⭐"""

    model, schema = orm.User, schemas.User

    @classmethod
    async def get_stats(cls, session: AsyncSession):
        """⭐"""
        query = (
            select(func.count("*").label("c"))
            .where(cls.model.marzban_username == None)
            )
        result = await session.execute(query)
        return result.scalar_one_or_none()

    @classmethod
    async def get_marzban_names(cls, session: AsyncSession):
        """⭐"""
        result = await session.execute(
            select(cls.model.marzban_username)
            .where(cls.model.marzban_username != None)
            )
        return result.scalars().all()

    @classmethod
    async def get_telegram_ids_by_marzban_names(cls, session, marzban_names):
        """⭐"""
        result = await session.execute(
            select(cls.model.telegram_id)
            .where(cls.model.marzban_username.in_(marzban_names))
            )
        return result.scalars().all()

    @classmethod
    async def get_telegram_ids_by_marzban_names_exp(cls, session, marzban_names):
        """⭐"""
        result = await session.execute(
            select(cls.model)
            .where(cls.model.marzban_username.in_(marzban_names))
            )
        for user in result.scalars().all():
            if len(user.payments) > 0:
                yield user.telegram_id

    @classmethod
    async def get_telegram_ids_by_marzban_names_exp_no_pay(cls, session, marzban_names):
        """⭐"""
        result = await session.execute(
            select(cls.model)
            .where(cls.model.marzban_username.in_(marzban_names))
            )
        for user in result.scalars().all():
            if len(user.payments) == 0:
                yield user.telegram_id


class PayPlanPricesRepository(Repository):
    """⭐"""

    model, schema, view = orm.PayPlanPrice, schemas.PayPlanPrice, schemas.PayPlanPriceView


class PayPlanPriceCurrenciesRepository(Repository):
    """⭐"""

    model, schema = orm.PayPlanPriceCurrency, schemas.PayPlanPriceCurrency


class PayPlansRepository(Repository):
    """⭐"""

    model, schema, view = orm.PayPlan, schemas.PayPlan, schemas.PayPlanView


class CouponsRepository(Repository):
    """⭐"""

    model, schema = orm.Coupon, schemas.Coupon


class UsageCouponsRepository(Repository):
    """⭐"""

    model, schema = orm.UsageCoupon, schemas.UsageCoupon

    @classmethod
    async def get_stats(cls, session: AsyncSession, coupon_id):
        """⭐"""
        query = (
            select(func.count("*").label("c"))
            .where(cls.model.coupon_id == coupon_id)
            )
        result = await session.execute(query)
        return result.scalar_one_or_none()

    @classmethod
    async def get_stats_grouped(cls, session: AsyncSession):
        """⭐"""
        query = (
            select(cls.model.coupon_id, func.count("*"))
            .group_by(cls.model.coupon_id)
            )
        result = await session.execute(query)
        return result.all()


class PaymentsRepository(Repository):
    """⭐"""

    model, schema = orm.Payment, schemas.Payment

    @classmethod
    async def get_stats(
        cls,
        session: AsyncSession,
        currency: str = settings.business.primary_currency
        ):
        """⭐"""
        dt_now = datetime.utcnow()
        dt_today = datetime.combine(dt_now, datetime.min.time())
        query_30d = (
            select(func.sum(cls.model.price).label("s"))
            .where(
                cls.model.ts_created >= dt_now - timedelta(days=30),
                cls.model.paid, cls.model.currency == currency
                )
            )
        query_1d = (
            select(func.sum(cls.model.price).label("s"))
            .where(
                cls.model.ts_created >= dt_today,
                cls.model.paid, cls.model.currency == currency
                )
            )
        query_total = (
            select(func.sum(cls.model.price).label("s"))
            .where(
                cls.model.paid, cls.model.currency == currency
                )
            )
        stats = {}
        for key, query in (
            ("payments_30d", query_30d),
            ("payments_1d", query_1d),
            ("payments_total", query_total)
        ):
            result = await session.execute(query)
            stats[key] = result.scalar_one_or_none() or 0
        return stats


class UserPaymentsRepository(Repository):
    """⭐"""

    model, schema = orm.UserPayment, schemas.UserPayment


class UserBalancesRepository(Repository):
    """⭐"""

    model, schema = orm.UserBalance, schemas.UserBalance


class ReferralPaymentsRepository(Repository):
    """⭐"""

    model, schema = orm.ReferralPayment, schemas.ReferralPayment

    @classmethod
    async def get_user_balance(
        cls,
        session: AsyncSession,
        user_id: int,
        currency: str = settings.business.primary_currency
        ):
        """⭐"""
        query = (
            select(func.sum(cls.model.amount).label("s"))
            .where(
                cls.model.user_id == user_id,
                cls.model.currency == currency
                )
            )
        result = await session.execute(query)
        return result.scalar_one_or_none()



# -*- coding: utf-8 -*-
from logging import getLogger
from pickle import dumps, loads
from typing import Any, Optional

from aiocache import Cache
from redis.typing import ExpiryT

from vpnflow.services._redis import create_client
from vpnflow.settings import settings

DEFAULT_TTL = settings.cache.default_ttl
logger = getLogger(__name__)
redis_url = settings.cache.redis_url.get_secret_value()

if redis_url:

    class CacheRepository:
        """⭐"""

        client = create_client(redis_url)
        # client.auto_close_connection_pool = True

        @classmethod
        async def get(cls, key: str) -> Optional[Any]:
            """⭐"""
            value = await cls.client.get(key)
            if value is not None:
                try:
                    value = loads(value)
                except TypeError as exc:
                    logger.warning(exc)
                    value = None
            logger.debug(f"Get from cache. Key: {key}. Value: {value}")
            return value

        @classmethod
        async def set(cls, key: str, value: Any, exp: Optional[ExpiryT] = None) -> bool:
            """⭐"""
            logger.debug(f"Set in cache. Key: {key}. Value: {value}. Exp: {exp}")
            await cls.client.set(key, dumps(value), ex=exp)
            return True

        @classmethod
        async def delete(cls, key: str) -> bool:
            """⭐"""
            logger.debug(f"Delete from cache. Key: {key}.")
            await cls.client.delete(key)
            return True

        @classmethod
        async def close(cls) -> bool:
            """⭐"""
            logger.debug("Close cache connection")
            await cls.client.aclose(close_connection_pool=True)
            return True

        @classmethod
        async def check_health(cls) -> bool:
            """⭐"""
            logger.info("Ping cache")
            try:
                r = await cls.client.ping()
                return r
            except Exception as exc:
                logger.error(exc)
                return False

else:

    class CacheRepository:
        """⭐"""

        client = Cache()

        @classmethod
        async def get(cls, key: str) -> Optional[Any]:
            """⭐"""
            value = await cls.client.get(key)
            logger.debug(f"Get from cache. Key: {key}. Value: {value}")
            return value

        @classmethod
        async def set(cls, key: str, value: Any, exp: Optional[ExpiryT] = None) -> bool:
            """⭐"""
            logger.debug(f"Set in cache. Key: {key}. Value: {value}. Exp: {exp}")
            await cls.client.set(key, value, ttl=exp)
            return True

        @classmethod
        async def delete(cls, key: str) -> bool:
            """⭐"""
            logger.debug(f"Delete from cache. Key: {key}.")
            await cls.client.delete(key)
            return True

        @classmethod
        async def close(cls) -> bool:
            """⭐"""
            return True

        @classmethod
        async def check_health(cls) -> bool:
            """⭐"""
            return True
